/**
  Copyright Red Hat, Inc. 2006

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING.  If not, write to the
  Free Software Foundation, Inc.,  675 Mass Ave, Cambridge, 
  MA 02139, USA.

  Author: Lon Hohberger <lhh at redhat.com>
 */
/**
  @file Quorum disk /proc/partition scanning functions
 */
#include <stdio.h>
#include <stdlib.h>
#include <disk.h>
#include <errno.h>
#include <sys/types.h>
#include <platform.h>
#include <stdlib.h>
#include <string.h>


int
check_device(char *device, char *label, int *ssz, quorum_header_t *qh,
	     int flags)
{
	int ret = -1;
	quorum_header_t qh_local;
	target_info_t disk;

	if (!qh)
		qh = &qh_local;

	ret = qdisk_validate(device);
	if (ret < 0) {
		perror("qdisk_verify");
		return -1;
	}

	ret = qdisk_open(device, &disk);
	if (ret < 0) {
		perror("qdisk_open");
		return -1;
	}

	if (ssz) 
		*ssz = disk.d_blksz;

	ret = -1;
	if (qdisk_read(&disk, OFFSET_HEADER, qh, sizeof(*qh)) == sizeof(*qh)) {
		swab_quorum_header_t(qh);
                if (qh->qh_magic == HEADER_MAGIC_NUMBER) {
			if (!label || !strcmp(qh->qh_cluster, label)) {
				ret = 0;
			}
                }
        }

	/* only flag now is 'strict device check'; i.e.,
	  "block size recorded must match kernel's reported size" */
	if (flags && qh->qh_version == VERSION_MAGIC_V2 &&
            disk.d_blksz != qh->qh_blksz) {
		ret = -1;
	}

	qdisk_close(&disk);

	return ret;
}


char *
state_str(disk_node_state_t s)
{
	switch (s) {
	case S_NONE:
		return "None";
	case S_EVICT:
		return "Evicted";
	case S_INIT:
		return "Initializing";
	case S_RUN:
		return "Running";
	case S_MASTER:
		return "Master";
	default:
		return "ILLEGAL";
	}
}


void
print_status_block(status_block_t *sb)
{
	if (sb->ps_state == S_NONE)
		return;
	printf("Status block for node %d\n", sb->ps_nodeid);
	printf("\tLast updated by node %d\n", sb->ps_updatenode);
	printf("\tLast updated on %s", ctime((time_t *)&sb->ps_timestamp));
	printf("\tState: %s\n", state_str(sb->ps_state));
	printf("\tFlags: %04x\n", sb->ps_flags);
	printf("\tScore: %d/%d\n", sb->ps_score, sb->ps_scoremax);
	printf("\tAverage Cycle speed: %d.%06d seconds\n", 
		sb->ps_ca_sec, sb->ps_ca_usec);
	printf("\tLast Cycle speed: %d.%06d seconds\n", 
		sb->ps_lc_sec, sb->ps_lc_usec);
	printf("\tIncarnation: %08x%08x\n",
		(int)(sb->ps_incarnation>>32&0xffffffff),
		(int)(sb->ps_incarnation&0xffffffff));

}


void
read_info(char *dev)
{
	target_info_t ti;
	int x;
	status_block_t sb;

	if (qdisk_open(dev, &ti) < 0) {
		printf("Could not read from %s: %s\n",
		       dev, strerror(errno));
		return;
	}

	for (x = 0; x < MAX_NODES_DISK; x++) {

		if (qdisk_read(&ti,
			       qdisk_nodeid_offset(x+1, ti.d_blksz),
			       &sb, sizeof(sb)) < 0) {
			printf("Error reading node ID block %d\n",
			       x+1);
			continue;
		}
		swab_status_block_t(&sb);
		print_status_block(&sb);
	}

	qdisk_close(&ti);
}


int
find_partitions(const char *partfile, const char *label,
	        char *devname, size_t devlen, int print)
{
	char line[4096];
	FILE *fp;
	int minor, major;
	unsigned long long blkcnt;
	char device[128];
	char realdev[256];
	quorum_header_t qh;
	int ssz;

	fp = fopen(partfile, "r");
	if (!fp)
		return -1;

	while (fgets(line, sizeof(line), fp) != NULL) {
		if (strlen(line) > 128 + (22) /* 5 + 5 + 11 + 1 */) {
			/*printf("Line too long!\n");*/
			continue;
		}

		/* This line is taken from 2.6.15.4's proc line */
		sscanf(line, "%4d %4d %10llu %s", &major, &minor,
		       &blkcnt, device);

		if (strlen(device)) {
			snprintf(realdev, sizeof(realdev),
				 "/dev/%s", device);

			/* If we're not "just printing", then 
			   then reject devices which don't match
			   the recorded sector size */
			if (check_device(realdev, (char *)label, &ssz,
					 &qh, !print) != 0)
				continue;

			if (print) {
				printf("%s:\n", realdev);
				printf("\tMagic:                %08x\n", qh.qh_magic);
				printf("\tLabel:                %s\n", qh.qh_cluster);
				printf("\tCreated:              %s",
				       ctime((time_t *)&qh.qh_timestamp));
				printf("\tHost:                 %s\n", qh.qh_updatehost);
				printf("\tKernel Sector Size:   %d\n", ssz);
				if (qh.qh_version == VERSION_MAGIC_V2) {
					printf("\tRecorded Sector Size: %d\n\n", (int)qh.qh_blksz);
					if (qh.qh_blksz != ssz) {
						printf("WARNING: Sector size mismatch: Header: %d  Kernel: %d\n",
							(int)qh.qh_blksz, ssz);
					}
				} else
					printf("\n");
			}

			if (print >= 2) {
				/* Print node stuff */
				read_info(realdev);
			}

			if (devname && devlen) {
				/* Got it */
				strncpy(devname, realdev, devlen);
				fclose(fp);
				return 0;
			}
		}
	}

	fclose(fp);

	if (print)
		/* No errors if we're just printing stuff */
		return 0;

	errno = ENOENT;
	return -1;
}
